#!/usr/bin/env python3
import hashlib
import os

from colorama import Fore, Style


class Warning:

    UNSTAGED_WARNING = (Fore.YELLOW + "\nYou have lint errors in unstaged files! "
        "Run python3 -m selective_linter -u to see those errors too" + Fore.RESET)
    BYPASS_WARNING = Fore.YELLOW + "Committing warnings or errors." + Fore.RESET


class LintError:

    def __init__(self, error):
        split_error = error.split(":")
        self.file = split_error[0]
        self.line = split_error[1]
        self.character = split_error[2]
        self.error_type = split_error[3]
        self.description = split_error[4]
        aux_length = len(":".join(split_error[:5])) + 1
        self.code = error[aux_length:]

    def __str__(self):
        where = self.file + ":" + self.line
        if "warning" in self.error_type:
            error_type = LintError._warning(self.description)
        elif "error" in self.error_type:
            error_type = LintError._error(self.description)
        return where + '\n' + error_type + self.code

    @staticmethod
    def _warning(warning):
        return Fore.YELLOW + "(warning)" + warning + ":" + Fore.RESET

    @staticmethod
    def _error(error):
        return Fore.RED + "(error)" + error + ":" + Fore.RESET


class ChangeVerifier:

    CACHE_NAME = '.lint_cache'

    def __init__(self, hunks):
        hashes = [hashlib.md5(hunk.encode()).hexdigest() for hunk in hunks]
        self.long_hash = "".join(hashes)
    
    def has_unchanged_cache(self):
        if not os.path.isfile(self.CACHE_NAME):
            return False
        long_hash = None
        with open(self.CACHE_NAME, 'r') as file:
            long_hash = file.read()
        return self.long_hash == long_hash

    def write_cache(self): 
        with open(self.CACHE_NAME, 'w+') as file:
            file.write(self.long_hash)

    def clear_cache(self):
        if os.path.isfile(self.CACHE_NAME):
            os.remove(self.CACHE_NAME)


#!/usr/bin/env python3
import os
import sh

class SwiftLint:

    def __init__(self, dir=None, files=[]):
        self.linter = sh.swiftlint.lint.bake() # pylint: disable=no-member
        self.files = files
        rootdir = dir or os.getcwd()
        os.chdir(os.path.abspath(rootdir))
        self.log = ""
        self.lint_errors = self._lint()

    def check_errors_against_diff(self, diff_lines):
        errors = set()
        for line in diff_lines:
            filename = line.split(':')[0]
            line_number = line.split(':')[1]
            potential_errors = self.lint_errors[filename]
            for error in potential_errors:
                location = ":".join([filename, line_number])
                if location in error:
                    errors.add(error)
        return errors

    def _lint(self):
        lint_errors = {}
        for filename in self.files:
            lint_errors[filename] = []
            try:
                lint_results = self.linter(filename)
                results = lint_results.stdout.decode('utf-8', 'ignore')
            except sh.ErrorReturnCode_2 as e:
                results = e.stdout.decode('utf-8', 'ignore')
                self.log = e.stderr.decode('utf-8', 'ignore')
            if results:
                results = results.split('\n')
                lint_errors[filename] += results
        return lint_errors




import os

import sh

SHEBANG = "#!/usr/bin/bash\n"
HOOK = """
PATH=$PATH:/usr/local/bin:/usr/local/sbin
skipswiftlint=$(git config --bool hooks.skipswiftlint)
if [ "$skipswiftlint" != "true" ]
then
    python3 -m selective_linter
    retVal=$?
    if [ $retVal != 0 ]
    then
    cat <<\EOF

You can disable this check by using:

    git config hooks.skipswiftlint true
EOF
fi
    exit $retVal
else
    exit 0
fi
"""
PRE_COMMIT = '/.git/hooks/pre-commit'
GITIGNORE = '/.gitignore'
LINT_CACHE = '.lint_cache'

class Installer:

    def __init__(self, dir):
        self.dir = dir 

    def is_repository(self):
        git_dir = self.dir + '/.git'
        if not os.path.isdir(git_dir):
            print("{} is not a valid git repository!".format(self.dir))
            return False
        return True

    def install(self): 
        pre_hook_dir = self.dir + PRE_COMMIT
        if os.path.exists(pre_hook_dir):
            with open(self.pre_hook_dir, 'r') as pre_commit_hook:
                if 'skipswiftlint' in pre_commit_hook.read():
                    return
            self._write_pre_commit_hook(append=True)
        else:
            self._write_pre_commit_hook(append=False)
        self._make_pre_commit_hook_executable()
        self._add_lint_cache_to_gitignore()

    @property
    def pre_hook_dir(self):
        return self.dir + PRE_COMMIT

    @property
    def gitignore_dir(self):
        return self.dir + GITIGNORE

    def _write_pre_commit_hook(self, append=False):
        if append:
            with open(self.pre_hook_dir, 'a') as pre_commit_hook:
                pre_commit_hook.write(HOOK)
        else: 
            with open(self.pre_hook_dir, 'w+') as pre_commit_hook:
                pre_commit_hook.write(SHEBANG)
                pre_commit_hook.write(HOOK)
        print("Wrote hook to {}".format(self.pre_hook_dir))
        self._sanitize_file(self.pre_hook_dir)
        
    def _sanitize_file(self, file):
        text = ''
        with open(file, 'r') as content:
            import pdb; pdb.set_trace()
            text = content.read()
        text.replace('\r', '')
        with open(file, 'w+') as content:
            content.write(text)

    def _make_pre_commit_hook_executable(self):
        sh.chmod('744', self.pre_hook_dir)
        print("Made hook executable")

    def _add_lint_cache_to_gitignore(self):
        should_add_to_gitignore = False
        with open(self.gitignore_dir, 'r') as gitignore:
            if LINT_CACHE not in gitignore.read():
                should_add_to_gitignore = True
        if should_add_to_gitignore:
            with open(self.gitignore_dir, 'a') as gitignore:
                gitignore.write('\n' + '# swiftlint')
                gitignore.write(LINT_CACHE + '\n')



#!/usr/bin/env python3
import argparse
import os

parser = argparse.ArgumentParser(description="""
    Call SwiftLint only on the lines changed relative to the base branch.
    """)
parser.add_argument('-d', '--dir', 
                    action='store', 
                    type=str,
                    default=os.getcwd())
parser.add_argument('-i', '--install',
                    action='store_true')
parser.set_defaults(install=False)
parser.add_argument('-u', '--unstaged',
                    action='store_true')
parser.set_defaults(unstaged=False)

running_as_script = False


def run():
    # Imports for running locally and running as module
    if running_as_script:
        from git_diff import GitDiff, GitHunk
        from linter import SwiftLint
        from output import LintError, ChangeVerifier, Warning
        from installer import Installer
    else:
        from selective_linter.git_diff import GitDiff
        from selective_linter.linter import SwiftLint
        from selective_linter.output import LintError, ChangeVerifier, Warning
        from selective_linter.installer import Installer

    # Parse command line arguments and install if necessary
    args = parser.parse_args()
    if args.install:
        installer = Installer(args.dir)
        if installer.is_repository():
            installer.install()
            return 0
        else: 
            print("Failed to install to directory")
            return 1

    # Process diffs and lint
    differ = GitDiff(args.dir)
    differ.diff(verbose=running_as_script or args.unstaged)
    linter = SwiftLint(dir=args.dir, files=differ.files_changed)
    errors = linter.check_errors_against_diff(differ.diff_lines)
    staged_errors = set([
        error for error in errors if error.split(":")[0] in differ.staged_files
    ])
    unstaged_errors = errors.difference(staged_errors)
    exit_code = 0
    cache = ChangeVerifier(differ.hunks)
    if errors:
        # If the user made the same commit twice, clear the cache and bypass this check
        if cache.has_unchanged_cache() and not args.unstaged:
            cache.clear_cache()
            if staged_errors:
                print(Warning.BYPASS_WARNING)
            else: 
                print(Warning.UNSTAGED_WARNING)
            return exit_code
        cache.write_cache()
        if staged_errors:
            for error in sorted(staged_errors):
                print(LintError(error))
            print("\nErrors found. Please fix errors and retry or make no changes and re-commit to bypass.")
            exit_code = 1
        else:
            if args.unstaged:
                print("Unstaged errors: ")
                for error in sorted(unstaged_errors):
                    print(LintError(error))
            else:
                print(Warning.UNSTAGED_WARNING)
    else:
        cache.clear_cache()
    return exit_code


if __name__ == "__main__":
    running_as_script = True
    run()



#!/usr/bin/env python3
import os
import re
import sh
import sys

# Constants

ANSI_ESCAPE_REGEX = r'\x1B[@-_][0-?]*[ -/]*[@-~]'
HUNK_HEADER_NEW_REGEX = r'(@@ -[0-9]+,[0-9]+ \+[0-9]+ @@)'
HUNK_HEADER_EXISTING_REGEX = r'(@@ -[0-9]+,[0-9]+ \+[0-9]+,[0-9]+ @@)'
BRANCH_REGEX = r'\[([A-Za-z0-9]+)]'

class GitHunk:

    def __init__(self, file, line_header, changes):
        self.file = file
        self.line_numbers = []
        self.lines_changed = []
        matches = re.search(r'\+([0-9]+)', line_header)
        line_number = int(matches.group(0)[1:])
        hunk = changes.split('\n')[1:]
        for line in hunk:
            if line.startswith('-'): 
                continue
            elif line.startswith('+'): 
                line = line[1:]
                self.line_numbers.append(line_number)
                self.lines_changed.append(line)
            line_number += 1

    def get_file_and_line_numbers(self):
        lines = []
        for line_number in self.line_numbers:
            lines.append("{}:{}".format(self.file, line_number))
        return lines

    def encode(self):
        return self.__str__().encode("utf-8")

    def __str__(self):
        lines_changed = ""
        for line_number, line in zip(self.line_numbers, self.lines_changed):
            lines_changed += "{}:{}: {}\n".format(self.file, line_number, line)
        return lines_changed


class GitDiff:

    def __init__(self, directory):
        self.parent_branch = "master"
        self.files_changed = []
        self.staged_files = []
        self.diff_lines = []
        self.hunks = []
        self._directory = os.path.abspath(directory)
        self._git = sh.git.bake(_cwd=directory)

    def diff(self, verbose=False):
        self.parent_branch = self._get_parent_branch()
        files_changed = [
            f for f in self._get_files_changed(staged_only=False) if os.path.isfile(f)
        ]
        staged_files = [
            f for f in self._get_files_changed(staged_only=True) if os.path.isfile(f)
        ]
        for filename in files_changed:
            self.hunks += self._get_diff_hunks(filename)
        self.files_changed = files_changed
        self.staged_files = staged_files
        self.diff_lines = self._get_diff_lines()
        if verbose:
            self._print_debug_info()

    def _print_debug_info(self):
        print("Using parent branch: {}".format(self.parent_branch))
        print("Total files changed: {}".format(len(self.files_changed)))
        print("Total files staged: {}".format(len(self.staged_files)))
        print("Total git hunks: {}".format(len(self.hunks)))
        if self.staged_files:
            print("Files staged: {}".format("\n\t".join(sorted(self.staged_files))))
        unstaged_files = sorted(set(self.files_changed).difference(set(self.staged_files)))
        if unstaged_files:
            print("Unstaged files: {}".format("\n\t".join(sorted(unstaged_files))))

    def _get_current_branch(self): 
        branch = self._git('rev-parse', '--abbrev-ref', 'HEAD')
        stdout = branch.stdout.decode('utf-8', 'ignore')
        return re.sub(ANSI_ESCAPE_REGEX, '', stdout)
    
    def _get_parent_branch(self):
        branch = self._git('show-branch', '-a')
        stdout = branch.stdout.decode('utf-8', 'ignore')
        branch_output = re.sub(ANSI_ESCAPE_REGEX, '', stdout)
        commits = branch_output.split('\n')
        current_branch = self._get_current_branch()
        commits_in_ancestor_branches = [
            commit for commit in commits if current_branch not in commit
        ]
        parent_commit = commits_in_ancestor_branches[0]
        return re.search(BRANCH_REGEX, parent_commit).group(1)

    def _get_files_changed(self, staged_only=False):
        if staged_only:
            files = self._git('--no-pager', 'diff', '--staged', '--name-only')
        else:
            files = self._git('--no-pager', 'diff', self.parent_branch, '--name-only')
        return [self._directory + '/' + filename 
                for filename in files.split('\n')
                if filename and filename.endswith(".swift") and os.path.isfile(filename)]
    
    def _get_diff_hunks(self, filename):
        hunks = []
        diff = self._git('--no-pager', 'diff', self.parent_branch, filename)
        stdout = diff.stdout.decode('utf-8', 'ignore')
        diff_output = re.sub(ANSI_ESCAPE_REGEX, '', stdout)
        regex = self._regex_for_diff_output(diff_output)
        captures = re.split(regex, diff_output)
        line_numbers = [group for (idx, group) in enumerate(captures[1:]) if idx % 2 == 0] 
        lines_changed = [group for (idx, group) in enumerate(captures[1:]) if idx % 2 == 1]
        for line, changes in zip(line_numbers, lines_changed):
            hunk = GitHunk(filename, line, changes)
            hunks.append(hunk)
        return hunks

    def _regex_for_diff_output(self, diff_output):
        if "new file" in diff_output:
             return HUNK_HEADER_NEW_REGEX
        return HUNK_HEADER_EXISTING_REGEX

    def _get_diff_lines(self):
        diff_lines = []
        for hunk in self.hunks:
            diff_lines += hunk.get_file_and_line_numbers()
        return diff_lines
        


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="selective_linter",
    version="0.1.0",
    author="Matt Boran",
    author_email="mattboran@gmail.com",
    description="A script to be used as a pre-commit git hook for projects written in Swift",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mattboran/SelectiveSwiftLinter",
    download_url="https://github.com/mattboran/SelectiveSwiftLinter/releases/download/0.1.0/selective_linter-0.1.0-py3-none-any.whl",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
    ],
    install_requires=[
        'colorama',
        'sh'
    ],
    python_requires='>3.0.0',
)
